/*
 * Copyright (C) 1998  Warren Jasper
 * All rights reserved.
 *
 */

/***************************************************************************
 *
 * c9513.c
 *
 ***************************************************************************/

#define MODULE
#define HIGH(x) (*((unsigned char*)&(x)+1))
#define LOW(x)  (*(unsigned char*)&(x))
#define GATEINTERVAL   1000

/* retrieve the CONFIG_* macros */
#include <linux/autoconf.h>
#if defined(CONFIG_MODVERSIONS) && !defined(MODVERSIONS)
#define MODVERSIONS
#endif

#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif

#include <linux/version.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/mm.h>
#include <linux/timer.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/signal.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include "ctr05.h"
#include "c9513.h"

/***************************************************************************
 *
 * Prototype of public and private functions.
 *
 ***************************************************************************/

static int ctr05_read(struct inode *iNode, struct file *filePtr, char *buf, int count);
static int ctr05_write(struct inode *iNode, struct file *filePtr, const char *buf, int count);
static int ctr05_open(struct inode *iNode, struct file *filePtr);
static void ctr05_close(struct inode *iNode, struct file *filePtr);
static int ctr05_ioctl(struct inode *iNode, struct file *filePtr, unsigned int cmd, LONG arg);
static void ctr05_TimerHandler(LONG);
static void ctr05_Interrupt();

int init_module(void);
void cleanup_module(void) ;

/***************************************************************************
 *
 * Global data.
 *
 ***************************************************************************/

static BoardRec BoardData;                    /* Board specific information      */
static CounterRec  Counter[NUM_COUNTERS];     /* Counter specific information    */
static DIO_ChanRec ChanDIO;                   /* DIO Channel specific information*/
static int MajorNumber = DEFAULT_MAJOR_DEV;   /* Major number compiled in        */
struct wait_queue  *adc0_wait;                /* wait semaphore                  */
struct timer_list TimerList = {NULL, NULL, 0, 0, ctr05_TimerHandler};


/***************************************************************************
 *
 *
 ***************************************************************************/

#if LINUX_VERSION_CODE < 0x020000
char kernel_version[] = UTS_RELEASE;        /* Required to by module             */
#endif

static struct file_operations ctr05_fops = {
	NULL,           /* ctr05_seek    */
	ctr05_read,     /* ctr05_read    */
	ctr05_write,    /* ctr05_write   */
	NULL,		/* ctr05_readdir */
	NULL,		/* ctr05_select  */
	ctr05_ioctl,	/* ctr05_ioctl   */
	NULL,
	ctr05_open,
	ctr05_close,
	NULL		/* fsync         */
};

#ifdef __cplusplus
extern "C" {
#endif


/***************************************************************************
 *
 * Loads driver. Called when "insmod ctr05.o" is invoked on the command line.
 *               The board is set to CTRO5_IRQ, 
 *
 ***************************************************************************/

int init_module(void) 
{
  int err;
  int i;
  BYTE bReg;

  /* Register as a device with kernel.  */

  if ((err = register_chrdev(MajorNumber, "ctr05", &ctr05_fops))) {
      printk("%s: Failure to load module. error %d\n", ADAPTER_ID, -err);
      return err;
  } 

  /* Register the region of iospace with kernel and get base address */

  if ( check_region(CTR05_BASE, BOARD_SIZE ) != 0 ) {  /* in use */
    printk("%s: Can't allocate region port address %#x\n", 
            ADAPTER_ID, CTR05_BASE );
    if (unregister_chrdev(MajorNumber, "ctr05") != 0) {
      printk("%s: unregister_chrdev() failed.\n", ADAPTER_ID);
    }
    return -ENODEV;
  } else {
    request_region( CTR05_BASE, BOARD_SIZE, "ctr05" );
  }

  /* Register interrupt handler. */
  
  if ( CTR05_IRQ >= 2 && CTR05_IRQ <= 7 ) {  /* only IRQ 2-7 are valid */
#if LINUX_VERSION_CODE > 0x020000
    if (request_irq(CTR05_IRQ, ctr05_Interrupt, SA_INTERRUPT, "ctr05", NULL)) {
#else
    if (request_irq(CTR05_IRQ, ctr05_Interrupt, SA_INTERRUPT, "ctr05")) {
#endif
      printk("%s: Can't request IRQ %d\n", ADAPTER_ID, CTR05_IRQ);
      if ( unregister_chrdev( MajorNumber, "ctr05" ) != 0 ) {  /* unregister device */
        printk( "%s: unregister_chrdev() failed.\n", ADAPTER_ID);
      }
      release_region( CTR05_BASE, BOARD_SIZE );       /* release region */
      return -(EIO);
    }
  }

  /* Initialize the 9513 chip */
  outb_p(MASTER_RESET, CMD_REG);      /* MASTER RESET */
  outb_p(MASTER_MODE_REG, CMD_REG);   /* Select Master Mode Register */
  bReg = TOD_DISABLED | MM_SRC1;
  outb_p(bReg, DATA_REG);
  bReg = 0xc0;        /* BCD Division, Disable increment, 8-Bit bus, FOUT On */
  outb_p(bReg, DATA_REG);

  BoardData.base = CTR05_BASE;
  BoardData.irq  = CTR05_IRQ;
  BoardData.mm = BCD_DIV | DISABLE_INCREMENT | TOD_DISABLED | MM_SRC1;
  BoardData.gateInterval=GATEINTERVAL;
  BoardData.useCounter = 0;

  for ( i = 0; i < NUM_COUNTERS; i++ ) {
    Counter[i].open = FALSE;
    Counter[i].count = 0;
    Counter[i].chan = i+1;
    Counter[i].cm = 0;
    Counter[i].mode = 0;
  }

  ChanDIO.open = FALSE;

  printk("%s: address=%#x IRQ=%d.",ADAPTER_ID, BoardData.base, BoardData.irq);
  printk(" 3/1/98 wjasper@tx.ncsu.edu\n");

  return 0;  
}

/***************************************************************************
 *                                                                         *
 * Remove driver. Called when "rmmod ctr05" is run on the command line.    *
 *                                                                         *
 ***************************************************************************/

void cleanup_module(void) 
{
  if (MOD_IN_USE) {
    printk("%s: device busy, remove delayed.\n", ADAPTER_ID);
    return;
  }

  release_region( BoardData.base, BOARD_SIZE );

  if ( CTR05_IRQ >= 2 && CTR05_IRQ <= 7 ) {  /* only IRQ 2-7 are valid */
#if LINUX_VERSION_CODE > 0x020000
    free_irq(CTR05_IRQ, NULL);
#else
    free_irq(CTR05_IRQ);
#endif
  }
  if (unregister_chrdev(MajorNumber, "ctr05") != 0) {
    printk("%s: cleanup_module failed.\n", ADAPTER_ID);
  } else {

  #ifdef DEBUG
      printk("%s: module removed.\n", ADAPTER_ID);
  #endif

  }    
}

#ifdef __cplusplus
}
#endif

/***************************************************************************
 *
 * open() service handler
 *
 ***************************************************************************/

static int ctr05_open(struct inode *iNode, struct file *filePtr)
{
  int minor = MINOR(iNode->i_rdev);

  /* 
     check if device is already open: only one process may read from a
     port at a time.  There is still the possibility of two processes 
     reading from two different channels messing things up. However,
     the overhead to check for this may not be worth it.
  */

  if ( minor == 0 ) {
    MOD_INC_USE_COUNT;
    ChanDIO.open = TRUE;                 /* The device is open */    
  }

  if ( minor > 0 && minor <= NUM_COUNTERS ) {
    --minor;                                 /* the array starts from 0 */
    if ( Counter[minor].open == TRUE ) {
      return -EBUSY;
    }

    MOD_INC_USE_COUNT;
    Counter[minor].open = TRUE;              /* The device is open */
    Counter[minor].mode = filePtr->f_flags;  /* set acquisition mode */

    /* if we are in frequency counting mode, set counter 5 busy */
    if ( Counter[minor].mode == CTR05_FREQUENCY ) {
      Counter[CTR5].open = TRUE;
      BoardData.useCounter++;
    }

    #ifdef DEBUG
        printk("%s: open(): minor %d mode %d\n", ADAPTER_ID, minor,
                Counter[minor].mode);
    #endif

  }
  return 0;
}

/***************************************************************************
 *
 * close() service handler
 *
 ***************************************************************************/

static void ctr05_close(struct inode *iNode, struct file *filePtr)
{
  int minor = MINOR(iNode->i_rdev);
  BYTE bReg;

  MOD_DEC_USE_COUNT;

  if ( minor == 0 ) {
    ChanDIO.open = FALSE;
  }

  if ( minor > 0 && minor <= NUM_COUNTERS ) {
    --minor;
    Counter[minor].open = FALSE;

    /* if we are in frequency counting mode, disarm counter */
    if ( Counter[minor].mode == CTR05_FREQUENCY ) {
      bReg = 1 << (minor-1);
      outb_p(DISARM | bReg, CMD_REG);         /* disarm  CTR minor */
      BoardData.useCounter--;
      if ( BoardData.useCounter == 0 ) {
        Counter[CTR5].open = FALSE;
        outb_p(DISARM | CTR5_BIT, CMD_REG);     /* disarm CTR5  */
      }
    }
  }

  #ifdef DEBUG
      printk("%s: close() of minor number %d.\n", ADAPTER_ID, minor);
  #endif
}

/***************************************************************************
 *
 * read() service function
 *
 ***************************************************************************/

static int ctr05_read(struct inode *iNode, struct file *filePtr, char *buf, int count)
{
  int minor = MINOR(iNode->i_rdev);
  int stat;
  int port;
  BYTE bReg;
  WORD data;


  #ifdef DEBUG
      printk("******* Read minor %d!!!!\n",minor);
  #endif

  if ( minor == 0 ) {
      bReg = inb(DIO_IN);
      put_fs_byte(bReg, buf);
      return 1;
  }

  if ( minor > 0 && minor <= NUM_COUNTERS ) {
    port = minor - 1;
    switch (Counter[port].mode) {
      case CTR05_COUNTER:    /* read the current count from the counter */

        /* Disarm and Save counters   */
	outb((DISARM_LATCH | 1<<(minor-1)), CMD_REG);

	/* Read the Hold Register */
        outb((minor | HOLD_REG), CMD_REG);       /* Fetch Hold Reg */
	data = inb_p(DATA_REG) & 0xff;           /* Low Byte */
        data |= (inb_p(DATA_REG) << 8) & 0xff00; /* High Byte */

        /* Check that data can be written to file. */
        if ((stat = verify_area(VERIFY_WRITE, buf, sizeof(WORD))) != 0) {
          printk("ctr05_read: Failed VERIFY_WRITE.\n");
          BoardData.busy = FALSE;
          return -1;
        }
	
	bReg = 1 << minor;
	if (inb_p(CMD_REG) & bReg) {
	  /* Counter overflow */
	  data = 0xFFFF;
	  printk("Counter overflow.\n");
	}

        /* Write data to user space */
        put_fs_word(data, (WORD*) buf);

        BoardData.busy = FALSE;
        break;
       
    case CTR05_FREQUENCY:

        /* Read the Hold Register */
        outb(HOLD_REG | minor, CMD_REG );
        data  = inb_p(DATA_REG) & 0xff;
        data |= (inb_p(DATA_REG) << 8) & 0xff00;

	#ifdef DEBUG
	   printk("read-Frequency read is %x.\n",data);
	#endif

	bReg = 1 << minor;
        if ( inb_p(CMD_REG) & bReg ) {
	  /* frequency overrun */
          data = 0xffff;
	  printk("Frequency overflow!\n");
	}
	/* Write data to user space */
	put_fs_word(data, (WORD*) buf);
        break;

      default:
        break;
    }
    return 1;     /* return 1 16 bit counter read */
  }
  return 0;
}

/***************************************************************************
 *
 * write() service function
 *
 ***************************************************************************/

static int ctr05_write(struct inode *iNode, struct file *filePtr, const char *buf, int count)
{
    int  minor = MINOR(iNode->i_rdev);
    BYTE bReg;

    #ifdef DEBUG
        printk("ctr05_write: minor = %d\n",minor);
    #endif

    if ( minor == 0 ) {  /* write to Digital Output Port */
      bReg = get_fs_byte(buf);
      outb( bReg, DIO_OUT);
      return 1;
    }

    /* write to LOAD register */
    if ( minor > 0 && minor <= NUM_COUNTERS ) { /* load counter */
      switch (Counter[minor-1].mode) {
        case CTR05_COUNTER:

	  /* set mode A for Count #minor */
	  outb_p(MODE_REG | minor, CMD_REG);
	  outb_p(COUNTUP, DATA_REG);
	  bReg = get_fs_word(buf) >> 8;
	  outb_p(bReg, DATA_REG);       /* SRC #minor */

	  /* Select the desired LOAD Register */
	  bReg = minor | LOAD_REG;
	  outb_p (bReg, CMD_REG);

	  /* Load the desired data into the register */
	  bReg=0;
	  outb_p (bReg, DATA_REG);    /* Low Byte */
	  outb_p (bReg, DATA_REG);    /* High Byte */

	  /* LOAD and ARM the counter now */
	  bReg = LOAD_ARM | (1 << (minor - 1));
	  outb_p (bReg, CMD_REG);
	  break;

        case CTR05_FREQUENCY:
	
	  bReg = 1 << (minor-1);
	  outb_p(DISARM | CTR5_BIT | bReg, CMD_REG);     /* disarm CTR5 & CTR minor */


	  /* below sets mode J for CTR5 */
	  outb_p(MODE_REG | CTR5, CMD_REG); /* Select counter 5                              */
	  outb_p( 0x62, DATA_REG);          /* Disable spec. gate, Reload from load or hold, */
                                            /* Count repeat, binary, Count down, TC toggled  */
	  outb_p(0xe, DATA_REG);            /* No gating, Count on rising edge, Freq4 input  */

	  outb_p(LOAD_REG | CTR5, CMD_REG); /* Setting up a variable duty pulse on CTR5      */
	  outb_p(5, DATA_REG);              /* the first "half" is 5 counts (from LOAD reg)  */
	  outb_p(0, DATA_REG);              /* the second "half" is "GateInterval" counts    */
                                            /* load a 5 count delay into CTR5 load reg       */

	  /* load gate interval into CTR5 hold reg */
	  outb_p(HOLD_REG | CTR5, CMD_REG); 
	  outb_p(BoardData.gateInterval, DATA_REG);
	  outb_p(BoardData.gateInterval >> 8, DATA_REG);

	  /* set Mode Q for Counter #minor */
	  outb_p(MODE_REG | minor, CMD_REG);
	  outb_p(COUNTUP | SPECIALGATE |RECYCLE, DATA_REG);
	  bReg = (get_fs_word(buf) | AHLGATE) >> 8;
	  outb_p(bReg , DATA_REG); 
	  
	  /* Load 0 into CTR minor's Load Reg. */
	  outb_p(LOAD_REG | minor, CMD_REG);
	  outb_p(0, DATA_REG);
	  outb_p(0, DATA_REG);
 
	  /*
	    NOTE: To setup a 32 bit counter, use CTR3 as the gate.
	    Setup CTR3 as CTR4 is setup above.  Then setup CTR4 to use Mode E with
	    it's input from SigSource (as for CTR5 above) and CTR3 as the gate.
	    Finally setup CTR5 using Mode B, but with its source as TCN-1 (the TC
	    output from CTR4 which will pulse each time CTR4 wraps) and also gated
	    by CTR3.  When done, CTR4 holds the low 16 bits and CTR5 holds the
	    high 16 bits.
	    */

	  bReg = 1 << (minor -1);
	  outb_p(LOAD | CTR5_BIT | bReg, CMD_REG);      /* Load CTR5 & CTR minor */
	  outb_p(CLEAR_OUTPUT | CTR5, CMD_REG);         /* set TC Low            */
	  outb_p (ARM | CTR5_BIT | bReg, CMD_REG);      /* Go ...                */

          /* Load 1 into CTR minor's Load Reg. */
          outb_p(LOAD_REG | minor, CMD_REG);
          outb_p(1, DATA_REG);
          outb_p(0, DATA_REG);
	
	  break;
      }
      return 1;
    }
    return 0;
}

/***************************************************************************
 *
 * iotctl() service handler
 *
 ***************************************************************************/

/* Note:
    Remember that FIOCLEX, FIONCLEX, FIONBIO, and FIOASYN are reserved ioctl cmd numbers
*/

static int ctr05_ioctl(struct inode *iNode, struct file *filePtr, unsigned int cmd, LONG arg)
{
  int minor = MINOR(iNode->i_rdev);
  BYTE *ptr = (BYTE*) &arg;
  BYTE bReg;
  int size = _IOC_SIZE(cmd);       /* the size bitfield in cmd */
  int err = 0;
  
  #ifdef DEBUG
     printk("ctr05_ioctl: minor = %d\n", minor);
  #endif

  /*
   * extract the type and number bitfields, and don't decode
   * wrong cmds;  return EINVAL before verify_area()
   */
  
  if (_IOC_TYPE(cmd) != IOCTL_MAGIC) return -EINVAL;
  if (_IOC_NR(cmd) > IOCTL_MAXNR)  return -EINVAL;
  
  if (_IOC_DIR(cmd) & _IOC_READ) {
    err = verify_area(VERIFY_WRITE, (void *)arg, size);
  } else if (_IOC_DIR(cmd) & _IOC_WRITE) {
    err = verify_area(VERIFY_READ, (void *)arg, size);
  }
  if (err) return err;

  if ( minor >= 1 && minor <= NUM_COUNTERS ) {
    switch (cmd) {
      case LOAD_MASTER_MODE_REG:
        outb_p(MASTER_MODE_REG, CMD_REG);        /* Select Master Mode Register */
        outb_p(*(ptr++), DATA_REG);
        outb_p(*ptr, DATA_REG);
        BoardData.mm = (WORD) arg;
        break;
      case LOAD_COUNTER_MODE_REG:
        outb_p(minor|MODE_REG, CMD_REG);         /* Select Counter Mode Register */
        outb_p(*(ptr++), DATA_REG);
        outb_p(*ptr, DATA_REG);
        Counter[minor-1].cm = (WORD) arg;
        break;
      case LOAD_ALARM_REG1:
        outb_p(ALARM_REG1, CMD_REG);
        outb_p(*(ptr++), DATA_REG);
        outb_p(*ptr, DATA_REG);
        break;
      case LOAD_ALARM_REG2:
        outb_p(ALARM_REG2, CMD_REG);
        outb_p(*(ptr++), DATA_REG);
        outb_p(*ptr, DATA_REG);
        break;
      case LOAD_CMD_REG:
        outb_p( (BYTE) arg, CMD_REG);
        break;
      case LOAD_DATA_REG:
	outb_p(minor|LOAD_REG, CMD_REG);
        outb_p(*(ptr++), DATA_REG);
        outb_p(*ptr, DATA_REG);
        break;
      case GET_MASTER_MODE_REG:
        put_fs_long((long) BoardData.mm, (long*) arg);
        break;
      case GET_COUNTER_MODE_REG:
        put_fs_long((long) Counter[minor-1].cm, (long*) arg);
        break;
      case GET_STATUS_REG:
        bReg = inb(CMD_REG);
        put_fs_long((long) bReg, (long*) arg);
        break;
      case SET_GATE_INTERVAL:
        BoardData.gateInterval = (WORD) arg;
        break;
    }  /* end switch */
  }  /* end if */
  return 0;
}

/***************************************************************************
 *
 *  Alarm handler to handle situations where an interrupt is missed.
 *
 ***************************************************************************/

static void ctr05_TimerHandler(LONG unused)
{
}

/***************************************************************************
 *
 * Interrupt handler used to service enhanced mode(interrupt) read().
 *
 ***************************************************************************/

static void ctr05_Interrupt(int irq)
{
  #ifdef DEBUG
      printk("Entering ctr05_ReadInterrupt(). irq = %d\n", CTR05_IRQ );
  #endif

}

